#include <ros/ros.h>
#include <body_msgs/Hands.h>
#include <body_msgs/Skeletons.h>
#include <body_msgs/SkeletonJoint.h>
#include <geometry_msgs/Point.h>
#include <teleop_msgs/body_angle.h>
#include "pcl/features/normal_3d.h"
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/radius_outlier_removal.h>
#include <pcl/filters/conditional_removal.h>

#define PI 3.14159265358979323846

sensor_msgs::PointCloud2 hand0; //mano izquerda
sensor_msgs::PointCloud2 hand1; //mano derecha
body_msgs::Skeletons skeleton; //esqueleto

//Se crea el publicador
ros::Publisher bodyAnglesPub;



//Se crea el mensaje a enviar
teleop_msgs::body_angle bodyAnglesMsg;

//Grado de confianza necesario para trabajar con un joint del esqueleto
double confidence = 0.05;


/*
 *Publica el mensaje
 */
void pubMensajes()
{
  //solo se publica si el esqueleto recibido tiene informacion
  if (skeleton.skeletons.size() != 0)
    bodyAnglesPub.publish(bodyAnglesMsg);
}

/*
 * Calcula el angulo formado entre la linea que une dos partes del cuerpo, y la linea que forman otras dos partes.
 */
float getAnguloSinVertice(geometry_msgs::Point joint1, geometry_msgs::Point joint2, geometry_msgs::Point joint3,
                          geometry_msgs::Point joint4)
{
  //Se crean dos vectores, cada uno representa la union entre dos partes del cuerpo
  double v1[3], v2[3];

  v1[0] = joint3.x - joint4.x;
  v1[1] = joint3.y - joint4.y;
  v1[2] = joint3.z - joint4.z;

  v2[0] = joint1.x - joint2.x;
  v2[1] = joint1.y - joint2.y;
  v2[2] = joint1.z - joint2.z;

  //Se calcula el tamanio de los vectores
  float v1_magnitud = sqrt((double)(v1[0] * v1[0] + v1[1] * v1[1] + v1[2] * v1[2]));
  float v2_magnitud = sqrt((double)(v2[0] * v2[0] + v2[1] * v2[1] + v2[2] * v2[2]));

  if (v1_magnitud != 0.0 && v2_magnitud != 0.0)
  {

    v1[0] = v1[0] * (1.0 / v1_magnitud);
    v1[1] = v1[1] * (1.0 / v1_magnitud);
    v1[2] = v1[2] * (1.0 / v1_magnitud);

    v2[0] = v2[0] * (1.0 / v2_magnitud);
    v2[1] = v2[1] * (1.0 / v2_magnitud);
    v2[2] = v2[2] * (1.0 / v2_magnitud);

    //Se calcula el angulo con el arcocoseno
    float theta = acos((double)(v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]));

    return theta;
  }
  else
    return -1; // Error code
}

/*
 *Calcula el angulo formado por una articulacion del cuerpo
 */
double getAnguloConVertice(geometry_msgs::Point joint1, geometry_msgs::Point joint2, geometry_msgs::Point joint3)
{

  double v1[3], v2[3]; // Se crean dos vectores, cada uno representa la union de una parte del cuerpo con la articulacion

  v1[0] = joint2.x - joint3.x;
  v1[1] = joint2.y - joint3.y;
  v1[2] = joint2.z - joint3.z;

  v2[0] = joint1.x - joint2.x;
  v2[1] = joint1.y - joint2.y;
  v2[2] = joint1.z - joint2.z;

  //Se calcula el tamanio de los vectores
  float v1_magnitud = sqrt((double)(v1[0] * v1[0] + v1[1] * v1[1] + v1[2] * v1[2]));
  float v2_magnitud = sqrt((double)(v2[0] * v2[0] + v2[1] * v2[1] + v2[2] * v2[2]));

  if (v1_magnitud != 0.0 && v2_magnitud != 0.0)
  {

    v1[0] = v1[0] * (1.0 / v1_magnitud);
    v1[1] = v1[1] * (1.0 / v1_magnitud);
    v1[2] = v1[2] * (1.0 / v1_magnitud);

    v2[0] = v2[0] * (1.0 / v2_magnitud);
    v2[1] = v2[1] * (1.0 / v2_magnitud);
    v2[2] = v2[2] * (1.0 / v2_magnitud);

    //Se calcula el angulo con el arcocoseno
    float theta = acos((double)(v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]));

    return theta;
  }
  else
    return -1; // Error code
}

/*
 * Devuelve la posicion relativa entre dos partes del cuerpo,  dice si una esta encima, debajo, a la derecha, a la izquierda
 * o delante de la otra.
 *                          ^
 *                         0|
 *                          |
 *                  3<______|_____>2
 *                         /|
 *                        / |
 *                      4/  |1
 *                          V
 */int obtenerPosicionRelativa(body_msgs::SkeletonJoint joint1, body_msgs::SkeletonJoint joint2)
{
  //Se calcula la distancias entra las coordenadas del codo y la mano
  double diff_x, diff_y, diff_z;
  diff_x = joint1.position.x - joint2.position.x;
  diff_y = joint1.position.y - joint2.position.y;
  diff_z = joint1.position.z - joint2.position.z;
  //                           | Y
  //                           |
  //                     ______|_____X
  //                          /|
  //                         / |
  //                       Z/  |
  //Si la mayor diferencia la tiene la x
  if (std::abs(diff_x) > std::abs(diff_y) && std::abs(diff_x) > std::abs(diff_z))
  {
    if (diff_x > 0)
      return 2;
    else
      return 3;
  }
  //Si la mayor diferencia la tiene la y
  else if (std::abs(diff_y) > std::abs(diff_x) && std::abs(diff_y) > std::abs(diff_z))
  {
    if (diff_y > 0)
      return 0;
    else
      return 1;
  }
  //Si la amyor diferencia la tiene la z
  else if (std::abs(diff_z) > std::abs(diff_x) && std::abs(diff_z) > std::abs(diff_y))
  {
    return 4;
  }
  return -1;
}

/*
 * Encuentra los indices correspondientes a los puntos que estan en los extremos de una nube de puntos.
 * vectorPuntosExtremos[0] -> indice del punto mas alto
 * vectorPuntosExtremos[1] -> indice del punto mas bajo
 * vectorPuntosExtremos[2] -> indice del punto mas a la izquierda
 * vectorPuntosExtremos[3] -> indice del punto mas a la derecha
 * vectorPuntosExtremos[4] -> indice del punto mas al frente
 */
void posicionDePuntosExtremos(pcl::PointCloud<pcl::PointXYZ> cloud, int vectorPuntosExtremos[5])
{
  double arribaMax = 2, abajoMax = -2, izquierdaMax = 2, derechaMax = -2, frenteMax = 2;
  //Inicializo el vector donde se almacenaran los indices
  for (int i = 0; i < 5; i++)
  {
    vectorPuntosExtremos[i] = 0;
  }

  //Se recorren todos los puntos de la nube
  for (uint i = 0; i < cloud.points.size(); i++)
  {
    //Si es el punto mas alto hasta ahora
    if (cloud.points[i].y < arribaMax)
    {
      arribaMax = cloud.points[i].y; //Se memoriza su posicion
      vectorPuntosExtremos[0] = i; //Se almacena su indice
    }
    //Si es el punto mas bajo hasta ahora
    else if (cloud.points[i].y > abajoMax)
    {
      abajoMax = cloud.points[i].y; //Se memoriza su posicion
      vectorPuntosExtremos[1] = i; //Se almacena su indice
    }
    //Si es el punto mas a la izquierda hasta ahora
    if (cloud.points[i].x < izquierdaMax)
    {
      izquierdaMax = cloud.points[i].x; //Se memoriza su posicion
      vectorPuntosExtremos[2] = i; //Se almacena su indice
    }
    //Si es el punto mas a la derecha hasta ahora
    else if (cloud.points[i].x > derechaMax)
    {
      derechaMax = cloud.points[i].x; //Se memoriza su posicion
      vectorPuntosExtremos[3] = i; //Se almacena su indice
    }
    //Si es el punto mas al frente hasta ahora
    if (cloud.points[i].z < frenteMax)
    {
      frenteMax = cloud.points[i].z; //Se memoriza su posicion
      vectorPuntosExtremos[4] = i; //Se almacena su indice
    }
  }
  //printf("arriba %f, abajo %f, izq %f, der %f, frente %f\n", arribaMax, abajoMax, izquierdaMax, derechaMax, frenteMax);
}

/*
 * Funcion auxiliar, para el brazo derecho, que devuelve true si la direccion indicada por pos no es la direccion donde se encuentra la munieca
 * con respecto al codo.
 * pos=                     ^
 *                         0|
 *                          |
 *                  3<______|_____>2
 *                         /|
 *                        / |
 *                      4/  |1
 *                          V
 */
bool noEsMuniecaD(int pos)
{
  switch (obtenerPosicionRelativa(skeleton.skeletons[0].right_elbow, skeleton.skeletons[0].right_hand))
  {
    case 0: //Si la mano esta encima del codo
      if (pos == 1 || pos == 0)
        return false;
      break;
    case 1: //Si la mano esta debajo del codo
      if (pos == 1 || pos == 0)
        return false;
      break;
    case 2: //Si la mano esta a la izquierda del codo
      if (pos == 3 || pos == 2)
        return false;
      break;
    case 3: //Si la mano esta a la derecha del codo
      if (pos == 3 || pos == 2)
        return false;
      break;
    case 4: //Si la mano esta frente al codo
      if (pos == 5)
        return false;
      break;
  }
  return true;
}

/*
 * Funcion auxiliar, para el brazo izquierdo, que devuelve true si la direccion indicada por pos no es la direccion donde se encuentra la munieca
 * con respecto al codo.
 * pos=                     ^
 *                         0|
 *                          |
 *                  3<______|_____>2
 *                         /|
 *                        / |
 *                      4/  |1
 *                          V
 */
bool noEsMuniecaI(int pos)
{
  switch (obtenerPosicionRelativa(skeleton.skeletons[0].left_elbow, skeleton.skeletons[0].left_hand))
  {
    case 0: //Si la mano esta encima del codo
      if (pos == 1 || pos == 0)
        return false;
      break;
    case 1: //Si la mano esta debajo del codo
      if (pos == 1 || pos == 0)
        return false;
      break;
    case 2: //Si la mano esta a la izquierda del codo
      if (pos == 3 || pos == 2)
        return false;
      break;
    case 3: //Si la mano esta a la derecha del codo
      if (pos == 3 || pos == 2)
        return false;
      break;
    case 4: //Si la mano esta frente al codo
      if (pos == 5)
        return false;
      break;
  }
  return true;
}

/*
 * Devuelve la posicion del dedo gordo derecho, para ello devuelve el punto de la mano mas alejado de su centro.
 */
pcl::PointXYZ obtenerPosicionDedoGordoD(pcl::PointCloud<pcl::PointXYZ> cloud, int indicesExtremos[5],
                                        Eigen::Vector4f centroMano)
{
  double mayorDistancia = 0; //Almacena la mayor distancia de un punto al centro
  int puntoConMayorDistancia = -1; //Almacena el indice del punto con mayor distancia al centro
  //Se crea un punto con la informacion del centro de la mano
  pcl::PointXYZ centroManoAux;
  centroManoAux.x = centroMano[0];
  centroManoAux.y = centroMano[1];
  centroManoAux.z = centroMano[2];
  //Recorro todos los puntos extremos, que se obtienen con la funcion posicionDePuntoExtremos()
  for (int i = 0; i < 5; i++)
  {
    //Calculo su distancia al centro de la mano
    double distancia = std::abs(pcl::euclideanDistance(cloud.points[indicesExtremos[i]], centroManoAux));
    //Si la distancia es mayor a la que se tiene almacenada y no es punto que este en la direccion de la munieca
    if (distancia > mayorDistancia && noEsMuniecaD(i))
    {
      mayorDistancia = distancia; //Se almacena la nueva mayor distancia
      puntoConMayorDistancia = i; //Se almacena el indice del punto con la nueva mayor distancia
    }
  }
  //Devuelve el punto correspondiente al dedo gordo
  return cloud.points[indicesExtremos[puntoConMayorDistancia]];
}

/*
 * Devuelve la posicion del dedo gordo izquierdo. La funcion es similar a la anterior.
 */
pcl::PointXYZ obtenerPosicionDedoGordoI(pcl::PointCloud<pcl::PointXYZ> cloud, int indicesExtremos[5],
                                        Eigen::Vector4f centroMano)
{
  double mayorDistancia = 0;
  int puntoConMayorDistancia = -1;
  pcl::PointXYZ centroManoAux;
  centroManoAux.x = centroMano[0];
  centroManoAux.y = centroMano[1];
  centroManoAux.z = centroMano[2];
  for (int i = 0; i < 5; i++)
  {
    double distancia = std::abs(pcl::euclideanDistance(cloud.points[indicesExtremos[i]], centroManoAux));
    if (distancia > mayorDistancia && noEsMuniecaI(i))
    {
      mayorDistancia = distancia;
      puntoConMayorDistancia = i;
    }
  }
  return cloud.points[indicesExtremos[puntoConMayorDistancia]];
}

/*
 * Funcion que calcula el angulo de la mano derecha, es decir, hacia donde esta mirando.
 */
double calcularAnguloManoDer(pcl::PointXYZ dedoGordo, Eigen::Vector4f centroMano)
{
  //Calculo la distancia en los 3 ejes entre el dedo gordo y el centro de la mano
  double distanciaX = dedoGordo.x - centroMano[0];
  double distanciaY = dedoGordo.y - centroMano[1];
  double distanciaZ = dedoGordo.z - centroMano[2];
  //Obtengo la posicion relativa de mano con respecto al codo
  int posicionCodo = obtenerPosicionRelativa(skeleton.skeletons[0].right_elbow, skeleton.skeletons[0].right_hand);
  switch (posicionCodo)
  {
    case 0: //Mano encima de codo
      //Si el el codo esta delante del hombro
      if (obtenerPosicionRelativa(skeleton.skeletons[0].right_shoulder, skeleton.skeletons[0].right_elbow) == 4)
      {
        if (std::abs(distanciaZ) > std::abs(distanciaX))
          return 0; //Dedo esta hacia atras
        if (distanciaX < 0)
          return -PI / 2; //Dedo esta hacia la  derecha
        return PI / 2; //Dedo esta hacia la izquierda
      }
      //Si el codo no esta delante del hombro
      else
      {
        if (std::abs(distanciaX) > std::abs(distanciaZ))
          return 0; //Dedo esta un lado
        if (distanciaZ < 0)
          return -PI / 2; //Dedo esta hacia delante
        return PI / 2; //Dedo esta hacia atras
      }
      break;
    case 1: //Mano debajo de codo
      if (std::abs(distanciaZ) > std::abs(distanciaX))
        return 0; //Dedo esta hacia delante
      if (distanciaX < 0)
        return -PI / 2; //Dedo esta hacia un lado
      return PI / 2; //Dedo esta hacia el otro lado
      break;
    case 2: //Mano a la izquierda del codo
      if (std::abs(distanciaY) > std::abs(distanciaZ))
        return 0; //Dedo esta hacia atra
      if (distanciaZ < 0)
        return PI / 2; //Dedo esta hacia abajo
      return -PI / 2; //Dedo esta hacia arriba
      break;
    case 3: //Mano a la derecha del codo
      if (std::abs(distanciaY) > std::abs(distanciaZ))
        return 0; //Dedo hacia arriba
      if (distanciaZ < 0)
        return -PI / 2; // Dedo hacia atras
      return PI / 2; //Dedo hacia delante
      break;
    case 4: //Mano delante del codo
      if (std::abs(distanciaY) > std::abs(distanciaX))
        return 0; //Dedo hacia arriba
      if (distanciaX < 0)
        return -PI / 2; //Dedo hacia un lado
      return PI / 2; //Dedo hacia el otro
      break;
  }
  return -1;
}

/*
 * Funcion que calcula el angulo de la mano izquierda, es decir, hacia donde esta mirando. Similar a la funciona anterior.
 */
double calcularAnguloManoIzq(pcl::PointXYZ dedoGordo, Eigen::Vector4f centroMano)
{
  double distanciaX = dedoGordo.x - centroMano[0];
  double distanciaY = dedoGordo.y - centroMano[1];
  double distanciaZ = dedoGordo.z - centroMano[2];
  int posicionCodo = obtenerPosicionRelativa(skeleton.skeletons[0].left_elbow, skeleton.skeletons[0].left_hand);

  switch (posicionCodo)
  {
    case 0:
      if (obtenerPosicionRelativa(skeleton.skeletons[0].left_shoulder, skeleton.skeletons[0].left_elbow) == 4)
      {
        if (std::abs(distanciaZ) > std::abs(distanciaX))
          return 0;
        if (distanciaX < 0)
          return -PI / 2;
        return PI / 2;
      }
      else
      {
        if (std::abs(distanciaX) > std::abs(distanciaZ))
          return 0;
        if (distanciaZ < 0)
          return PI / 2;
        return -PI / 2;
      }
      break;
    case 1:
      if (std::abs(distanciaZ) > std::abs(distanciaX))
        return 0;
      if (distanciaX < 0)
        return -PI / 2;
      return PI / 2;
      break;
    case 2:
      if (std::abs(distanciaY) > std::abs(distanciaZ))
        return 0;
      if (distanciaZ < 0)
        return PI / 2;
      return -PI / 2;
      break;
    case 3:
      if (std::abs(distanciaY) > std::abs(distanciaZ))
        return 0;
      if (distanciaZ < 0)
        return -PI / 2;
      return PI / 2;
      break;
    case 4:
      if (std::abs(distanciaY) > std::abs(distanciaX))
        return 0;
      if (distanciaX < 0)
        return -PI / 2;
      return PI / 2;
      break;
  }
  return -1;
}

/*
 * Manejador que despierta cuando llega la nube de puntos de la mano derecha
 */
void hand1Handle(const sensor_msgs::PointCloud2ConstPtr &hand1Ptr)
{
  hand1 = *hand1Ptr; //Almaceno la nube de puntos de la mano derecha
  //Si la nube no tiene informacion, se termina la funcion
  if (hand1.width == 0)
    return;
  //Si el esqueleto no tiene informacion, se termina la funcion
  if (skeleton.skeletons.size() == 0)
    return;
  //Si no hay informacion sobre las partes del cuerpo necesarias, se termina.
  if (skeleton.skeletons[0].right_shoulder.confidence < confidence
      || skeleton.skeletons[0].right_elbow.confidence < confidence
      || skeleton.skeletons[0].right_hand.confidence < confidence)
    return;

  //  Se filtra la nube de puntos para reducir el numero de puntos y acelerar su procesamieto
  sensor_msgs::PointCloud2::Ptr cloud_tipo_ros(new sensor_msgs::PointCloud2());
  pcl::VoxelGrid<sensor_msgs::PointCloud2> sor;
  sor.setInputCloud(hand1Ptr);
  sor.setLeafSize(0.01f, 0.01f, 0.01f);
  sor.filter(*cloud_tipo_ros);
  pcl::PointCloud<pcl::PointXYZ> cloud;
  pcl::fromROSMsg(*cloud_tipo_ros, cloud);
  // Fin filtro. Reducir numero de puntos

  Eigen::Vector4f centroMano;
  pcl::compute3DCentroid(cloud, centroMano); //Se obtiene el centro de la mano

  // Se filtra para eliminar puntos de la munieca
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sin_munieca(new pcl::PointCloud<pcl::PointXYZ>);
  //Se calcula la posicion relativa de la mano con respecto al codo
  int posicionCodo = obtenerPosicionRelativa(skeleton.skeletons[0].right_elbow, skeleton.skeletons[0].right_hand);
  std::string ejeDeLaMunieca;
  double margen = 0.0;
  double distanciaAlCentro = 0.03;
  pcl::ComparisonOps::CompareOp tipoComparacion = pcl::ComparisonOps::LT;
  switch (posicionCodo)
  {
    case 0: //Si la mano esta encima del codo, la nube se cortara por abajo
      ejeDeLaMunieca = "y";
      margen = centroMano[1] + distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::LT;
      break;
    case 1: //si la mano esta bajo el codo, la nube se cortara por arriba
      ejeDeLaMunieca = "y";
      margen = centroMano[1] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
    case 2: //Si la mano esta a la izquierda del codo, la nube se cortara por la derecha
      ejeDeLaMunieca = "x";
      margen = centroMano[0] + distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::LT;
      break;
    case 3: //Si la mano esta a la derecha del codo, la nube se cortara por la izquierda
      ejeDeLaMunieca = "x";
      margen = centroMano[0] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
    case 4: //Si la mano esta en frente del codo, la nube se cortara por detras
      ejeDeLaMunieca = "z";
      margen = centroMano[2] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
  }
  //Se forma la condicion para cortar la nube
  pcl::ConditionBase<pcl::PointXYZ>::Ptr condicion(new pcl::ConditionAnd<pcl::PointXYZ>());
  condicion->addComparison(
      pcl::FieldComparison<pcl::PointXYZ>::ConstPtr(
          new pcl::FieldComparison<pcl::PointXYZ>(ejeDeLaMunieca, tipoComparacion, margen)));
  // Se construye el filtro
  pcl::ConditionalRemoval<pcl::PointXYZ> condrem(condicion);
  condrem.setInputCloud(cloud.makeShared());
  condrem.setKeepOrganized(true);
  // Se aplica el filtro
  condrem.filter(*cloud_sin_munieca);
  // Fin filtro. Eliminar puntos munieca

  int extremos[5]; //Almacenara los indices del punto mas alto, mas bajo, mas a la izquierda... dentro de la nube de puntos
  posicionDePuntosExtremos(*cloud_sin_munieca, extremos);
  double tamanioY = cloud_sin_munieca->points[extremos[0]].y - cloud_sin_munieca->points[extremos[1]].y;
  double tamanioX = cloud_sin_munieca->points[extremos[2]].x - cloud_sin_munieca->points[extremos[3]].x;
  double radioMano;
  if (std::abs(tamanioX) > std::abs(tamanioY))
    radioMano = std::abs(tamanioX) / 3.5;
  else
    radioMano = std::abs(tamanioY) / 3.5;
  if (radioMano < 0.035)
    radioMano = 0.035;

  //Se aplica un filtro para eliminar los puntos que tienen pocos vecinos ( y quitar el dedo gordo)
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sin_dedo(new pcl::PointCloud<pcl::PointXYZ>);
  pcl::RadiusOutlierRemoval<pcl::PointXYZ> outrem;
  // Se construye el filtro
  outrem.setInputCloud(cloud.makeShared());
  outrem.setRadiusSearch(radioMano);
  outrem.setMinNeighborsInRadius(60);
  // Se aplica el filtro
  outrem.filter(*cloud_sin_dedo);
  //Fin filtro. Eliminar puntos con pocos vecinos (dedo gordo)

  //Se obtiene el centro de la mano sin dedo gordo, ni munieca
  Eigen::Vector4f centroManoFiltrada;
  pcl::compute3DCentroid(*cloud_sin_dedo, centroManoFiltrada);
  //Se obtiene la posicion del dedo gordo
  pcl::PointXYZ dedoGordo;
  if (radioMano == 0.035) //Si el radio de la mano es pequeño significa que el dedo gordo esta orientado en el eje Z
  {
    dedoGordo.x = centroManoFiltrada[0];
    dedoGordo.y = centroManoFiltrada[1];
    if (centroManoFiltrada[2] - cloud_sin_munieca->points[extremos[4]].z > 0.04) //Si el dedo esta hacia el frente
      dedoGordo.z = centroManoFiltrada[2] - 0.7;
    else
      //Si esta hacia atras
      dedoGordo.z = centroManoFiltrada[2] + 0.7;
  }
  else
    dedoGordo = obtenerPosicionDedoGordoD(*cloud_sin_munieca, extremos, centroManoFiltrada);
  //A partir del dedo gordo y del centro de la mano sin dedo gordo, ni munieca, se calcula la posicion de la mano
  bodyAnglesMsg.RHandRotationAngle = calcularAnguloManoDer(dedoGordo, centroManoFiltrada);
}

/*
 * Manejador que despierta cuando llega la nube de puntos de la mano izquierda
 */
void hand0Handle(const sensor_msgs::PointCloud2ConstPtr &hand0Ptr)
{
  hand0 = *hand0Ptr; //Almaceno la nube de puntos de la mano izquierda
  if (hand0.width == 0)
    return;

  if (skeleton.skeletons.size() == 0)
    return;

  if (skeleton.skeletons[0].left_shoulder.confidence < confidence
      || skeleton.skeletons[0].left_elbow.confidence < confidence
      || skeleton.skeletons[0].left_hand.confidence < confidence)
    return;
  //  Se filtra la nube de puntos para reducir el numero de puntos y acelerar su procesamieto
  sensor_msgs::PointCloud2::Ptr cloud_tipo_ros(new sensor_msgs::PointCloud2());
  pcl::VoxelGrid<sensor_msgs::PointCloud2> sor;
  sor.setInputCloud(hand0Ptr);
  sor.setLeafSize(0.01f, 0.01f, 0.01f);
  sor.filter(*cloud_tipo_ros);
  pcl::PointCloud<pcl::PointXYZ> cloud;
  pcl::fromROSMsg(*cloud_tipo_ros, cloud);
  //

  //Se obtiene el centro de la mano
  Eigen::Vector4f centroMano;
  pcl::compute3DCentroid(cloud, centroMano);

  // Se filtra para eliminar puntos de la munieca
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sin_munieca(new pcl::PointCloud<pcl::PointXYZ>);
  int posicionCodo = obtenerPosicionRelativa(skeleton.skeletons[0].left_elbow, skeleton.skeletons[0].left_hand);
  std::string ejeDeLaMunieca;
  double margen = 0.0;
  double distanciaAlCentro = 0.03;
  pcl::ComparisonOps::CompareOp tipoComparacion = pcl::ComparisonOps::LT;
  switch (posicionCodo)
  {
    case 0:
      ejeDeLaMunieca = "y";
      margen = centroMano[1] + distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::LT;
      break;
    case 1:
      ejeDeLaMunieca = "y";
      margen = centroMano[1] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
    case 2:
      ejeDeLaMunieca = "x";
      margen = centroMano[0] + distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::LT;
      break;
    case 3:
      ejeDeLaMunieca = "x";
      margen = centroMano[0] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
    case 4:
      ejeDeLaMunieca = "z";
      margen = centroMano[2] - distanciaAlCentro;
      tipoComparacion = pcl::ComparisonOps::GT;
      break;
  }
  pcl::ConditionBase<pcl::PointXYZ>::Ptr condicion(new pcl::ConditionAnd<pcl::PointXYZ>());
  condicion->addComparison(
      pcl::FieldComparison<pcl::PointXYZ>::ConstPtr(
          new pcl::FieldComparison<pcl::PointXYZ>(ejeDeLaMunieca, tipoComparacion, margen)));
  pcl::ConditionalRemoval<pcl::PointXYZ> condrem(condicion);
  condrem.setInputCloud(cloud.makeShared());
  condrem.setKeepOrganized(true);
  condrem.filter(*cloud_sin_munieca);
  // Fin filtro

  int extremos[5]; //Almacenara los indices del punto mas alto, mas bajo, mas a la izquierda... dentro de la nube de puntos
  posicionDePuntosExtremos(*cloud_sin_munieca, extremos);
  double tamanioY = cloud_sin_munieca->points[extremos[0]].y - cloud_sin_munieca->points[extremos[1]].y;
  double tamanioX = cloud_sin_munieca->points[extremos[2]].x - cloud_sin_munieca->points[extremos[3]].x;
  double radioMano;
  if (std::abs(tamanioX) > std::abs(tamanioY))
    radioMano = std::abs(tamanioX) / 3.5;
  else
    radioMano = std::abs(tamanioY) / 3.5;
  if (radioMano < 0.035)
    radioMano = 0.035;
  //Se aplica un filtro para eliminar los puntos que tienen pocos vecinos ( y quitar el dedo gordo)
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_sin_dedo(new pcl::PointCloud<pcl::PointXYZ>);
  pcl::RadiusOutlierRemoval<pcl::PointXYZ> outrem;
  outrem.setInputCloud(cloud.makeShared());
  outrem.setRadiusSearch(radioMano);
  outrem.setMinNeighborsInRadius(60);
  outrem.filter(*cloud_sin_dedo);
  //Fin filtro

  //Se obtiene el centro de la mano sin dedo gordo, ni munieca
  Eigen::Vector4f centroManoFiltrada;
  pcl::compute3DCentroid(*cloud_sin_dedo, centroManoFiltrada);
  //Se obtiene la posicion del dedo gordo
  pcl::PointXYZ dedoGordo;
  if (radioMano == 0.035) //Si el radio de la mano es pequeño significa que el dedo gordo esta orientado en el eje Z
  {
    dedoGordo.x = centroManoFiltrada[0];
    dedoGordo.y = centroManoFiltrada[1];
    if (centroManoFiltrada[2] - cloud_sin_munieca->points[extremos[4]].z > 0.04) //Si el dedo esta hacia el frente
      dedoGordo.z = centroManoFiltrada[2] - 0.7;
    else
      //Si esta hacia atras
      dedoGordo.z = centroManoFiltrada[2] + 0.7;
  }
  else
    dedoGordo = obtenerPosicionDedoGordoI(*cloud_sin_munieca, extremos, centroManoFiltrada);
  //A partir del dedo gordo y del centro de la mano sin dedo gordo, ni munieca, se calcula la posicion de la mano
  bodyAnglesMsg.LHandRotationAngle = calcularAnguloManoIzq(dedoGordo, centroManoFiltrada);
  pubMensajes();
}

/*
 * Funcion que calcula los angulos que forma el esqueleto recibido para crear
 * el mensaje a enviar. Antes de cada calculo se comprueba que se ha recibido informacion de todas las
 * partes del cuerpo necesarias para llevarlo a cabo.
 */
void calcular_angulos()
{
  if (skeleton.skeletons[0].right_hip.confidence > confidence && skeleton.skeletons[0].left_hip.confidence > confidence
      && skeleton.skeletons[0].right_shoulder.confidence > confidence
      && skeleton.skeletons[0].right_elbow.confidence > confidence)
//Es 0 si el brazo esta cerrado, 90 si esta abierto.
    bodyAnglesMsg.RShoulderOpeningAngle = getAnguloSinVertice(skeleton.skeletons[0].right_hip.position,
                                                              skeleton.skeletons[0].left_hip.position,
                                                              skeleton.skeletons[0].right_shoulder.position,
                                                              skeleton.skeletons[0].right_elbow.position) - PI / 2;
  if (skeleton.skeletons[0].right_hip.confidence > confidence
      && skeleton.skeletons[0].right_shoulder.confidence > confidence
      && skeleton.skeletons[0].right_elbow.confidence > confidence)
//Es 0 si el brazo esta pegado al cuerpo, 90 si se gira hacia delante, 180 si se estira hacia arriba
    bodyAnglesMsg.RShoulderRotationAngle = (getAnguloConVertice(skeleton.skeletons[0].right_hip.position,
                                                               skeleton.skeletons[0].right_shoulder.position,
                                                               skeleton.skeletons[0].right_elbow.position) - 8 * PI / 9)
        * -1;

  if (skeleton.skeletons[0].right_shoulder.confidence > confidence
      && skeleton.skeletons[0].right_elbow.confidence > confidence
      && skeleton.skeletons[0].right_hand.confidence > confidence)
    //0 si esta estirado, 90 si doblado
    bodyAnglesMsg.RElbowOpeningAngle = getAnguloConVertice(skeleton.skeletons[0].right_shoulder.position,
                                                            skeleton.skeletons[0].right_elbow.position,
                                                            skeleton.skeletons[0].right_hand.position);

  if (skeleton.skeletons[0].right_shoulder.confidence > confidence
      && skeleton.skeletons[0].left_shoulder.confidence > confidence
      && skeleton.skeletons[0].right_elbow.confidence > confidence
      && skeleton.skeletons[0].right_hand.confidence > confidence
      && skeleton.skeletons[0].right_hip.confidence > confidence)
  {
    //Depende de la apertura del hombro, si esta abierto sera: 90 si el codo esta doblado hacia arriba, 0 hacia el frente, -90 hacia abajo.
    //                                   si esta cerrado : 0 hacia el cuerpo,   90 hacia el frente, 180 hacia el exterior
    double rotacionCodoDDesdeHombros = getAnguloSinVertice(skeleton.skeletons[0].right_shoulder.position,
                                                           skeleton.skeletons[0].left_shoulder.position,
                                                           skeleton.skeletons[0].right_elbow.position,
                                                           skeleton.skeletons[0].right_hand.position);
    double rotacionCodoDDesdeCaderas = getAnguloSinVertice(skeleton.skeletons[0].right_shoulder.position,
                                                           skeleton.skeletons[0].right_hip.position,
                                                           skeleton.skeletons[0].right_elbow.position,
                                                           skeleton.skeletons[0].right_hand.position);
    bodyAnglesMsg.RElbowRotationAngle = rotacionCodoDDesdeCaderas * (bodyAnglesMsg.RShoulderOpeningAngle / (PI / 2))
        + rotacionCodoDDesdeHombros * (1 - bodyAnglesMsg.RShoulderOpeningAngle / (PI / 2));
  }



  if (skeleton.skeletons[0].right_shoulder.confidence > confidence)
    //Coordenadas del hombro derecho
    bodyAnglesMsg.RShoulderPosition = skeleton.skeletons[0].right_shoulder.position;

  if (skeleton.skeletons[0].right_hand.confidence > confidence)
     //Coordenadas de la mano derecho
     bodyAnglesMsg.RHandPosition = skeleton.skeletons[0].right_hand.position;

  if (skeleton.skeletons[0].head.confidence > confidence)
      //Coordenadas de la cabeza
      bodyAnglesMsg.Head = skeleton.skeletons[0].head.position;

  //Aqui empieza el calculo del lado izquierdo del cuerpo, identico al lado derecho.

  if (skeleton.skeletons[0].left_hip.confidence > confidence && skeleton.skeletons[0].right_hip.confidence > confidence
      && skeleton.skeletons[0].left_shoulder.confidence > confidence
      && skeleton.skeletons[0].left_elbow.confidence > confidence)
    bodyAnglesMsg.LShoulderOpeningAngle = getAnguloSinVertice(skeleton.skeletons[0].left_hip.position,
                                                              skeleton.skeletons[0].right_hip.position,
                                                              skeleton.skeletons[0].left_shoulder.position,
                                                              skeleton.skeletons[0].left_elbow.position) - PI / 2;

  if (skeleton.skeletons[0].left_hip.confidence > confidence
      && skeleton.skeletons[0].left_shoulder.confidence > confidence
      && skeleton.skeletons[0].left_elbow.confidence > confidence)
    bodyAnglesMsg.LShoulderRotationAngle = (getAnguloConVertice(skeleton.skeletons[0].left_hip.position,
                                                               skeleton.skeletons[0].left_shoulder.position,
                                                               skeleton.skeletons[0].left_elbow.position) - 8 * PI / 9)
        * -1;

  if (skeleton.skeletons[0].left_shoulder.confidence > confidence
      && skeleton.skeletons[0].left_elbow.confidence > confidence
      && skeleton.skeletons[0].left_hand.confidence > confidence)
    bodyAnglesMsg.LElbowOpeningAngle = getAnguloConVertice(skeleton.skeletons[0].left_shoulder.position,
                                                            skeleton.skeletons[0].left_elbow.position,
                                                            skeleton.skeletons[0].left_hand.position);

  if (skeleton.skeletons[0].left_shoulder.confidence > confidence
      && skeleton.skeletons[0].right_shoulder.confidence > confidence
      && skeleton.skeletons[0].left_elbow.confidence > confidence
      && skeleton.skeletons[0].left_hand.confidence > confidence
      && skeleton.skeletons[0].left_hip.confidence > confidence)
  {
    double rotacionCodoIDesdeHombros = getAnguloSinVertice(skeleton.skeletons[0].left_shoulder.position,
                                                           skeleton.skeletons[0].right_shoulder.position,
                                                           skeleton.skeletons[0].left_elbow.position,
                                                           skeleton.skeletons[0].left_hand.position);
    double rotacionCodoIDesdeCaderas = getAnguloSinVertice(skeleton.skeletons[0].left_shoulder.position,
                                                           skeleton.skeletons[0].left_hip.position,
                                                           skeleton.skeletons[0].left_elbow.position,
                                                           skeleton.skeletons[0].left_hand.position);
    bodyAnglesMsg.LElbowRotationAngle = rotacionCodoIDesdeCaderas * (bodyAnglesMsg.LShoulderOpeningAngle / (PI / 2))
        + rotacionCodoIDesdeHombros * (1 - bodyAnglesMsg.LShoulderOpeningAngle / (PI / 2));
  }


  if (skeleton.skeletons[0].left_shoulder.confidence > confidence)
    bodyAnglesMsg.LShoulderPosition = skeleton.skeletons[0].left_shoulder.position;

  if (skeleton.skeletons[0].left_hand.confidence > confidence)
      bodyAnglesMsg.LHandPosition = skeleton.skeletons[0].left_hand.position;

}

/*
 * Manejador que se despierta cuando se recibe un esqueleto
 */
void skelHandle(const body_msgs::SkeletonsConstPtr &skelsPtr)
{

  skeleton = *skelsPtr;
  calcular_angulos(); //llamada a la función que forma gran parte del mensaje
}



int main(int argc, char** argv)
{
  ros::init(argc, argv, "body_capture");
  ros::NodeHandle n;

  //Se anuncian los topic que se van a publicar
  bodyAnglesPub = n.advertise<teleop_msgs::body_angle>("bodyAngle", 1);


  //Se realiza la subscripcion a los distintos topics
  ros::Subscriber hand0Sub, hand1Sub, skelSub;
  hand0Sub = n.subscribe("/hand0_fullcloud", 1, hand0Handle); //mano izquierda
  hand1Sub = n.subscribe("/hand1_fullcloud", 1, hand1Handle); //mano derecha
  skelSub = n.subscribe("/skeletons", 1, skelHandle); //Esqueleto

  //Comienza el bucle principal
    ros::spin();

}
